/* TRASHERNET Stand-alone Ethernet Controller with SPI
 *
 * Copyright (c) 2016 Markus Koch <markus@notsyncing.net>
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#define DT_DRV_COMPAT notsyncing_trashernet

#define LOG_MODULE_NAME eth_trashernet
#define LOG_LEVEL CONFIG_ETHERNET_LOG_LEVEL

// TODO: Load this stuff from DT
#define TRASHERNET_BASE 0x82000000
#define TRASHERNET_SR (*((volatile uint32_t *) (TRASHERNET_BASE + 0x00)))
#define TRASHERNET_DR (*((volatile uint8_t *) (TRASHERNET_BASE + 0x04)))

#define TRASHERNET_SR_LINK_UP (1 << 0)
#define TRASHERNET_SR_RX_AVAILABLE (1 << 1)
#define TRASHERNET_SR_TX_NOTEMPTY (1 << 2)

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <string.h>
#include <errno.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/net/net_pkt.h>
#include <zephyr/net/net_if.h>
#include <zephyr/net/ethernet.h>

#define ETH_TRASHERNET_RX_THREAD_STACK_SIZE 800
#define ETH_TRASHERNET_RX_THREAD_PRIO 2

struct eth_trashernet_config {
	int32_t timeout;
};

struct eth_trashernet_runtime {
	struct net_if *iface;
	K_KERNEL_STACK_MEMBER(thread_stack, ETH_TRASHERNET_RX_THREAD_STACK_SIZE);
	struct k_thread thread;
	struct k_sem int_sem;
	bool iface_initialized : 1;
	uint8_t mac_address[6];
};

static void eth_trashernet_set_data(const struct device *dev, uint8_t *buffer, uint16_t length)
{
	uint16_t i;
	for (i=0; i < length; ++i) {
		TRASHERNET_DR = buffer[i];
	}
}

static int eth_trashernet_tx(const struct device *dev, struct net_pkt *pkt)
{
	struct eth_trashernet_runtime *context = dev->data;
	uint16_t len = net_pkt_get_len(pkt);
	struct net_buf *frag;

	LOG_DBG("TX %s: pkt %p (len %u)", dev->name, pkt, len);

	while (TRASHERNET_SR & TRASHERNET_SR_TX_NOTEMPTY); // FIXME: To be fixed in HDL: We currently do not support TX packet separation, so we have to wait for the TX buffer to be completely empty before adding the next packet.

	for (frag = pkt->frags; frag; frag = frag->frags) {
		eth_trashernet_set_data(dev, frag->data, frag->len);
	}
	TRASHERNET_SR = TRASHERNET_SR_TX_NOTEMPTY; // Frame complete, TX now

	//LOG_DBG("%s: Tx successful", dev->name);

	return 0;
}

static void eth_trashernet_get_data(const struct device *dev, uint8_t *buffer, uint16_t length)
{
	for (int x = 0; x < length; ++x) {
		buffer[x] = TRASHERNET_DR;
	}
}

static int eth_trashernet_rx(const struct device *dev)
{
	struct eth_trashernet_runtime *context = dev->data;
	const struct eth_trashernet_config *config = dev->config;
	uint16_t frm_len;
	uint16_t lengthfr;
	
	struct net_buf *pkt_buf;
	struct net_pkt *pkt;
	
	frm_len = TRASHERNET_SR >> 16;
	TRASHERNET_SR = (1 << 1); // Confirm that we've read the length
	if (!frm_len) {
		LOG_ERR("%s: Receive called without data available!", dev->name);
		return 0;
	}
	//LOG_DBG("RX Data %d", frm_len);
	
	pkt = net_pkt_rx_alloc_with_buffer(context->iface, frm_len,
					   AF_UNSPEC, 0, K_MSEC(config->timeout));
	if (!pkt) {
		LOG_ERR("%s: Could not allocate rx buffer", dev->name);
		return 0;
	}

	pkt_buf = pkt->buffer;
	lengthfr = frm_len;
	
	do {
		size_t frag_len;
		uint8_t *data_ptr;
		size_t spi_frame_len;

		data_ptr = pkt_buf->data;

		/* Review the space available for the new frag */
		frag_len = net_buf_tailroom(pkt_buf);

		if (frm_len > frag_len) {
			spi_frame_len = frag_len;
		} else {
			spi_frame_len = frm_len;
		}

		eth_trashernet_get_data(dev, data_ptr, spi_frame_len);

		net_buf_add(pkt_buf, spi_frame_len);

		/* One fragment has been written via SPI */
		frm_len -= spi_frame_len;
		pkt_buf = pkt_buf->frags;
	} while (frm_len > 0);

	net_pkt_set_iface(pkt, context->iface);

	LOG_DBG("%s: Received packet of length %u", dev->name, lengthfr);
	if (net_recv_data(net_pkt_iface(pkt), pkt) < 0) {
		net_pkt_unref(pkt);
	}

	return 0;
}

static void eth_trashernet_rx_thread(void *p1, void *p2, void *p3)
{
	int status_last = 0;
	int status = 0;
	ARG_UNUSED(p2);
	ARG_UNUSED(p3);

	const struct device *dev = p1;
	struct eth_trashernet_runtime *context = dev->data;

	while (true) {
		//k_sem_take(&context->int_sem, K_FOREVER); // Set to wait for IRQ
		//k_sem_give(&context->int_sem); // Set in IRQ

		status = TRASHERNET_SR;
		if ((status ^ status_last) & TRASHERNET_SR_LINK_UP) { // Link change
			if (status & TRASHERNET_SR_LINK_UP) {
				LOG_INF("%s: Link up", dev->name);
				net_eth_carrier_on(context->iface);
			} else {
				LOG_INF("%s: Link down", dev->name);

				if (context->iface_initialized) {
					net_eth_carrier_off(context->iface);
				}
			}
		}
		if (status & TRASHERNET_SR_RX_AVAILABLE) { // New RX data
			eth_trashernet_rx(dev);
		} else {
			k_msleep(10); // We poll
		}
		status_last = status;
	}
}

static enum ethernet_hw_caps eth_trashernet_get_capabilities(const struct device *dev)
{
	ARG_UNUSED(dev);

	return ETHERNET_LINK_10BASE_T
#if defined(CONFIG_NET_VLAN)
		| ETHERNET_HW_VLAN
#endif
		;
}

static void eth_trashernet_iface_init(struct net_if *iface)
{
	const struct device *dev = net_if_get_device(iface);
	struct eth_trashernet_runtime *context = dev->data;

	net_if_set_link_addr(iface, context->mac_address,
			     sizeof(context->mac_address),
			     NET_LINK_ETHERNET);

	if (context->iface == NULL) {
		context->iface = iface;
	}

	ethernet_init(iface);

	net_if_carrier_off(iface);
	context->iface_initialized = true;
}

static const struct ethernet_api api_funcs = {
	.iface_api.init		= eth_trashernet_iface_init,

	.get_capabilities	= eth_trashernet_get_capabilities,
	.send			= eth_trashernet_tx,
};

static int eth_trashernet_init(const struct device *dev)
{
	const struct eth_trashernet_config *config = dev->config;
	struct eth_trashernet_runtime *context = dev->data;

	/* Start interruption-poll thread */
	k_thread_create(&context->thread, context->thread_stack,
			ETH_TRASHERNET_RX_THREAD_STACK_SIZE,
			eth_trashernet_rx_thread,
			(void *)dev, NULL, NULL,
			K_PRIO_COOP(ETH_TRASHERNET_RX_THREAD_PRIO),
			0, K_NO_WAIT);

	LOG_INF("%s: Initialized", dev->name);
	return 0;
}

#define TRASHERNET_DEFINE(inst)                                                                      \
	static struct eth_trashernet_runtime eth_trashernet_runtime_##inst = {                         \
		.mac_address = DT_INST_PROP(inst, local_mac_address),                              \
		.int_sem = Z_SEM_INITIALIZER((eth_trashernet_runtime_##inst).int_sem, 0, UINT_MAX),  \
	};                                                                                         \
                                                                                                   \
	static const struct eth_trashernet_config eth_trashernet_config_##inst = { \
		.timeout = 500, \
	};                                                                                         \
                                                                                                   \
	ETH_NET_DEVICE_DT_INST_DEFINE(inst, eth_trashernet_init, NULL, &eth_trashernet_runtime_##inst, \
				      &eth_trashernet_config_##inst, CONFIG_ETH_INIT_PRIORITY,       \
				      &api_funcs, NET_ETH_MTU);

DT_INST_FOREACH_STATUS_OKAY(TRASHERNET_DEFINE);
